(*
   PPRSL - pretty printer for RSL specifications
   Copyright (C) 2001 Andrzej Wasowski, wasowski@data.pl

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*)
{
	val commentNestLevel = ref 0;   
	val dumbNestLevel = ref 0;   
	fun accept() = true;
	fun mathc c = (print ( "$\\"^c^"$" ); accept())
	fun normt t = (print t; accept())
	
	fun escape l = let fun escaper [] = ""						
			     | escaper (c::tail) = (case c of
						       #"_" => "\\_"
						     | #"\\" => "\\\\"
						     | _    => implode[c])^escaper(tail)
                       in escaper(explode l) end
						  
	
	fun keywd k = (print ( "`\\textbf{"^escape( k )^"}`" ); accept())
	fun ident i = (print ( "`"^escape( i )^"`" ); accept())

	local 
		val keywords = ["axiom", "class", "end", "false", "true",
		"type", "chaos", "scheme", "extend", "with", "value", "Bool", "Int",
		"Real", "Nat", "Unit", "Char", "Text", "set", "infset", "in", "rng", "map",
		"card", "abs", "real", "pre", "post", "as", "if", "then", "else", "elsif",
		"case", "let", "in", "hd", "tl", "len", "elems", "inds",
		"dom", "rng", "forall","of","object","while"]
		fun lkeyword( s, [] ) = false
		  | lkeyword( s, h::tail) = if s=h then true else lkeyword(s,tail)
	
	in fun is_keyword s = lkeyword( s, keywords ) end

	fun greek s = mathc( String.extract( s, 1, NONE ) )
	
	fun name  n = ((if is_keyword n then (keywd n) 
				      else (ident n)); 
		       accept())

}

rule Initial = 
        parse "all"     { mathc "forall" }
	| "exists"      { mathc "exists" }
	| "inter"	{ mathc "cap" }
        | "union"       { mathc "cup" }	
	| "isin"	{ mathc "in"	}
	| "~isin"       { mathc "notin" }
	| "-list"       { normt "$^*$"  }
	| "-inflist"	{ normt "$^\\omega$" }
	| "is"          { mathc "equiv" }
	| [`a`-`z``A`-`Z`]([`_``a`-`z``A`-`Z``0`-`9`])*([`'`])*  { name(Lexing.getLexeme lexbuf) }
	| "/*"          { commentNestLevel := 1; print "/*"; Comment lexbuf; accept() }
	| "/**"		{ dumbNestLevel := 1; Dumb lexbuf; accept() }
	| "/*`"         { print "`"; EscapeComment lexbuf; accept() }
	| "><"          { mathc "times"  }
	| "||"		{ mathc "or" }
	| "**"		{ mathc "Rpower" }
	| "/\\"		{ mathc "Rand" }
	| "\\/"         { mathc "Ror" }
	| "=>"          { mathc "Rimplies" }
	| "~="          { mathc "neq" }
	| ">="		{ mathc "ge" }
	| "<="		{ mathc "le" }
	| "-m->"	{ mathc "Rmap" }
	| "-m~->"       { mathc "Rpmap" }
	| "+>"          { mathc "mapsto" }
	| "!!"          { mathc "dagger" }
	| "-~->"	{ mathc "Rpinto" }
	| "->"		{ mathc "Rinto" }
	| "<->"         { mathc "leftrightarrow" } 
	| "_"           { normt "`\\underline{\\phantom{w}}`" }
	| "#"		{ normt "$^\\circ$" }
	| ":-"		{ mathc "cdot" }
	| "<<"          { mathc "subset" }
	| "<<="         { mathc "subseteq" }
	| ">>"          { mathc "supset" }
	| ">>="         { mathc "supseteq" }
	| "<..>"        { mathc "langle\\rangle" }
	| "<."          { mathc "langle" }
        | ".>"          { mathc "rangle" }
	| "-\\"         { mathc "lambda" }
	| "^"           { mathc "hat{\\phantom{o}}" }
	| "..."         { mathc "dots"  }
	| eof           { false }
	| "`"           { print "`"; PassDirectly lexbuf }
	| "`Sigma"	{ mathc "Sigma" }
	| "`sigma"	{ mathc "sigma" }
	| "`Gamma"	{ mathc "Gamma" }
	| "`alpha"	{ greek (getLexeme lexbuf) }
	| "`beta"	{ greek (getLexeme lexbuf) }
	| "`gamma"	{ greek (getLexeme lexbuf) }
	| "`delta"	{ greek (getLexeme lexbuf) }
	| "`epsilon"	{ greek (getLexeme lexbuf) }
	| "`zeta"	{ greek (getLexeme lexbuf) }
	| "`eta"	{ greek (getLexeme lexbuf) }
	| "`theta"	{ greek (getLexeme lexbuf) }
	| "`iota"	{ greek (getLexeme lexbuf) }
	| "`kappa"	{ greek (getLexeme lexbuf) }
	| "`mu"		{ greek (getLexeme lexbuf) }
	| "`nu"		{ greek (getLexeme lexbuf) }
	| "`xi"		{ greek (getLexeme lexbuf) }
	| "`pi"		{ greek (getLexeme lexbuf) }
	| "`rho"	{ greek (getLexeme lexbuf) }
	| "`sigma"	{ greek (getLexeme lexbuf) }
	| "`tau"	{ greek (getLexeme lexbuf) }
	| "`upsilon"	{ greek (getLexeme lexbuf) }
	| "`phi"	{ greek (getLexeme lexbuf) }
	| "`chi"	{ greek (getLexeme lexbuf) }
	| "`psi"	{ greek (getLexeme lexbuf) }
	| "`omega"	{ greek (getLexeme lexbuf) }
	| "`Gamma"	{ greek (getLexeme lexbuf) }
	| "`Delta"	{ greek (getLexeme lexbuf) }
	| "`Theta"	{ greek (getLexeme lexbuf) }
	| "`Lambda"	{ greek (getLexeme lexbuf) }
	| "`Xi"		{ greek (getLexeme lexbuf) }
	| "`Pi"		{ greek (getLexeme lexbuf) }
	| "`Sigma"	{ greek (getLexeme lexbuf) }
	| "`Upsilon"	{ greek (getLexeme lexbuf) }
	| "`Phi"	{ greek (getLexeme lexbuf) }
	| "`Psi"	{ greek (getLexeme lexbuf) }
	| "`Omega"	{ greek (getLexeme lexbuf) }
	| "$"           { print "$"; PassMathMode lexbuf }
	| "\t"          { print "        "; accept() }
	| _		{ print (Lexing.getLexeme lexbuf); accept() }
   
   and PassDirectly =
        parse "`"       { print "`"; accept() }
	| [^`\```_`]    { print (Lexing.getLexeme lexbuf); PassDirectly lexbuf }
	| "_"           { print "\\_"; PassDirectly lexbuf }
	| _             { print (Lexing.getLexeme lexbuf); PassDirectly lexbuf }
 
   and PassMathMode =
        parse "$"       { print "$"; accept() }
	| "\\$"         { print "\\$"; PassMathMode lexbuf }
	| [^`$`]        { print (Lexing.getLexeme lexbuf); PassMathMode lexbuf }
	| _             { print (Lexing.getLexeme lexbuf); PassMathMode lexbuf }



   and Comment =
	parse "*/"      { print "*/"; 
			  if !commentNestLevel = 1  then ()                                              
	                  else (commentNestLevel:=(!commentNestLevel-1);
				Comment lexbuf) 
			}
	| "/*"          { print "/*"; 
	                  commentNestLevel := !commentNestLevel+1;
			  Comment lexbuf 
			}                                 
	| "/**"		{ dumbNestLevel:=(!dumbNestLevel+1);
	                  Dumb lexbuf }
	| "/*`"		{ print "`"; EscapeComment lexbuf; Comment lexbuf}

	| [^`/``*``\n`` ``\t``\``]+
			{ print ("`"^(getLexeme lexbuf)^"`"); Comment lexbuf }                                  
	| `\t`          { print "        "; Comment lexbuf }
	| _             { print (getLexeme lexbuf); Comment lexbuf }    

   and Dumb =
   	parse "**/"	{ if !dumbNestLevel = 1 then ()
			  else (dumbNestLevel:=(!dumbNestLevel-1);
			        Dumb lexbuf)
			}
	| "/**"         { dumbNestLevel := !dumbNestLevel+1; Dumb lexbuf }
	| [^`/``*`]+	{ Dumb lexbuf }
	| _		{ Dumb lexbuf }
			
   and EscapeComment =
   	parse "`*/"	{ print "`" }
	| [^`/``*``\``]+{ print (getLexeme lexbuf); EscapeComment lexbuf }
	| _		{ print (getLexeme lexbuf); EscapeComment lexbuf }

;
